from numpy import arange, exp
import textwrap

order = 2    # Order of Taylor expansions, can be zero or a +ve integer
step = 0.1   # step size must divide (x_max - x_min) and x_min
x_min = -600
x_max = 600

def divides_by(num, divisor):
    div = num / divisor
    return abs(div-round(div)) < 1e-6
assert(divides_by(x_max - x_min, step))

knots = arange(x_min, x_max+step, step)

if x_min <= 0:
    # Make sure zero is in the list (assumed by the C code!)
    assert(divides_by(x_min, step))
    assert(abs(knots[-x_min/step]) < 1e-6)
    knots[-x_min/step] = 0

#print knots

wrapper = textwrap.TextWrapper(initial_indent="    ",
        subsequent_indent="    ", width=78)

def lstr(x):
    """format number as a string with "stupidly high" precision.
    But don't do silly things with smallish integers."""
    if (abs(x) < 1000) and (round(x) == x):
        return str(x)
    else:
        return '%15.15e' % x # Is 15 sensible?

str_vals = ' \\\n'.join(wrapper.wrap(', '.join([lstr(exp(x)) for x in knots])))

# This code assumes that the knots include zero
recipricol_step = 1./step
bin_offset = x_min*recipricol_step

## Example:
#val = 10.102
#bin = round(val*recipricol_step)
#idx = bin - bin_offset
#bin_val = bin*step
#xx = val - bin_val
#print knots[idx]
#print bin_val
#approx_exp = exp(knots[idx])*(1 + xx + xx*xx/2 + xx*xx*xx/2/3 +
#        xx*xx*xx*xx/2/3/4)
#print approx_exp
#print exp(val)

code = '''/* for lrint and INFINITY */
#define _ISOC9X_SOURCE  1
#define _ISOC99_SOURCE  1
#include "math.h"\n\n'''
code += '''#define X_MIN %s\n''' % lstr(x_min)
code += '''#define X_MAX %s\n''' % lstr(x_max)
code += '''#define STEP %s\n''' % lstr(step)
code += '''#define BIN_OFFSET %s\n''' % lstr(bin_offset)
code += '''#define RECIPRICOL_STEP %s\n''' % lstr(recipricol_step)
code += '\nstatic const double exp_table[] = { \\\n%s};' % str_vals
code +='''

void
texp(double* inputs, double* answer, int nn)
/* answer array must be allocated before calling! */
{
    int bin, idx;
    double val;'''
if order > 0:
    code += '''
    double bin_val, xx;'''
if order > 1:
    code += '''
    double %s;''' % ', '.join(['t%d' % i for i in range(2, order + 1)])
code +='''
    int i;

#ifdef _OPENMP
#pragma omp parallel for private(val, bin, idx, bin_val, xx'''
if order > 1:
    code += ''.join([', t%d' % i for i in range(2, order+1)])
code += ''')
#endif
    for (i = 0; i < nn; ++i) {
        val = inputs[i];
        if (val <= X_MIN) {
            answer[i] = 0;
        } else if (val >= X_MAX) {
            answer[i] = INFINITY;
        } else {
            bin = lrint(val*RECIPRICOL_STEP);
            idx = bin - BIN_OFFSET;'''
if order > 0:
    code += '''
            bin_val = bin*STEP;
            xx = val - bin_val; '''
if order > 1:
    code += '''
            t2 = xx*xx/2;'''
for i in range(3, order+1):
    code += '''
            t%d = t%d*xx/%d;''' % (i, i-1, i)
if order == 0:
    code += '''
            answer[i] = exp_table[idx];'''
elif order == 1:
    code += '''
            answer[i] = exp_table[idx]*(1 + xx);'''
elif order > 1:
    code += '''
            answer[i] = exp_table[idx]*(1 + xx + %s);''' % \
            ' + '.join(['t%d' % i for i in range(2, order+1)])
else:
    assert(False)
code += '''
        }
    }
}
'''

fid = open('texp.c', 'w')
fid.write(code)
fid.close()

